"""MSA simulator"""

import _Sailfish
import warnings
import pathlib
from typing import Dict, Optional, List
from .protocol import SimProtocol
from .distributions import PoissonDistribution
from .msa import Msa
from .constants import MODEL_CODES, SIMULATION_TYPE


# TODO delete one of this (I think the above if not used)
class BlockTreePython:
    '''
    Used to contain the events on a multiple branches (entire tree).
    '''
    def __init__(self, branch_block_dict: Dict[str, _Sailfish.Block]):
        self._branch_block_dict = branch_block_dict
        # dictionary of {str: List of blocks}
        self._branch_block_dict_python = {i: x for i, x in branch_block_dict.items()}
    
    def _get_Sailfish_blocks(self) -> Dict[str, _Sailfish.Block]:
        return self._branch_block_dict
    
    def get_branches_str(self) -> str:
        return {i: self._branch_block_dict[i].print_tree() for i in list(self._branch_block_dict.keys())}
    
    def get_specific_branch(self, branch: str) -> str:
        if not branch in self._branch_block_dict_python:
            raise ValueError(f"branch not in the _branch_block, aviable branches are: {list(self._branch_block_dict_python.keys())}")
        return self._branch_block_dict[branch].print_tree()
    
    def print_branches(self) -> str:
        for i in list(self._branch_block_dict.keys()):
            print(f"branch = {i}")
            print(self._branch_block_dict[i].print_tree())
    
    def block_list(self)  -> List:
        if not branch in self._branch_block_dict_python:
            raise ValueError(f"branch not in the _branch_block, aviable branches are: {list(self._branch_block_dict_python.keys())}")
        return self._branch_block_dict_python[branch]

class Simulator:
    """Simulate MSAs based on SimProtocol"""
    
    def __init__(
        self, 
        simProtocol: Optional[SimProtocol] = None,
        simulation_type: Optional[SIMULATION_TYPE] = None
    ):
        if not simProtocol:
            warnings.warn("initalized a simulator without simProtocol -> using a default protocol with Tree = '(A:0.01,B:0.5,C:0.03);' and root length of 100")
            # default simulation values
            possion = PoissonDistribution(10, 100)
            simProtocol = SimProtocol(tree="(A:0.01,B:0.5,C:0.03);")
            simProtocol.set_insertion_length_distributions(possion)
            simProtocol.set_deletion_length_distributions(possion)
            simProtocol.set_insertion_rates(0.05)
            simProtocol.set_deletion_rates(0.05)
            simProtocol.set_sequence_size(100)
            simProtocol.set_min_sequence_size(1)

        # verify sim_protocol
        if self._verify_sim_protocol(simProtocol):
            self._simProtocol = simProtocol
            if simulation_type == SIMULATION_TYPE.PROTEIN:
                self._simulator = _Sailfish.AminoSimulator(self._simProtocol._sim)
            else:
                self._simulator = _Sailfish.NucleotideSimulator(self._simProtocol._sim)
        else:
            raise ValueError("failed to verify simProtocol")
        
        if not simulation_type:
            warnings.warn("simulation type not provided -> running indel only simulation")
            simulation_type = SIMULATION_TYPE.NOSUBS
        
        if simulation_type == SIMULATION_TYPE.PROTEIN:
            self._alphabet = _Sailfish.alphabetCode.AMINOACID
        elif simulation_type == SIMULATION_TYPE.DNA:
            self._alphabet = _Sailfish.alphabetCode.NUCLEOTIDE
        elif simulation_type == SIMULATION_TYPE.NOSUBS:
            self._alphabet = _Sailfish.alphabetCode.NULLCODE
        else:
            raise ValueError(f"unknown simulation type, please provde one of the following: {[e.name for e in SIMULATION_TYPE]}")
        
        self._simulation_type = simulation_type
        self._is_sub_model_init = False
    
    def _verify_sim_protocol(self, simProtocol) -> bool:
        if not simProtocol.get_tree():
            raise ValueError("protocol miss tree, please provide when initalizing the simProtocol")
        if not simProtocol.get_sequence_size() or simProtocol.get_sequence_size() == 0:
            raise ValueError("protocol miss root length, please provide -> simProtocol.set_sequence_size(int)")
        if not simProtocol.get_insertion_length_distribution(0):
            raise ValueError("protocol miss insertion length distribution, please provide -> simProtocol.set_insertion_length_distributions(float)")
        if not simProtocol.get_deletion_length_distribution(0):
            raise ValueError("protocol miss deletion length distribution, please provide -> simProtocol.set_deletion_length_distributions(float)")
        if simProtocol.get_insertion_rate(0) < 0:
            raise ValueError(f"please provide a non zero value for insertion rate, provided value of: {simProtocol.get_insertion_rate(0)} -> simProtocol.set_insertion_rate(float)")
        if simProtocol.get_deletion_rate(0) < 0:
            raise ValueError(f"please provide a non zero value for deletion rate, provided value of: {simProtocol.get_deletion_rate(0)} -> simProtocol.set_deletion_rate(float)")
        return True
    
    def reset_sim(self):
        # TODO, complete
        pass

    def _create_site_rate_model(
        self,
        gamma_alpha: float = 1.0,
        gamma_categories: int = 1,
        invariant_proportion: float = 0.0,
        site_rate_correlation: float = 0.0
    ) -> tuple[List[float], List[float], List[List[float]]]:
        """
        Create rate categories and probabilities for the site rate model.
        
        Args:
            gamma_alpha: Alpha parameter for gamma distribution
            gamma_categories: Number of gamma rate categories
            invariant_proportion: Proportion of invariant sites (0 to <1)
            site_rate_correlation: Correlation between adjacent sites (0 to <1)
                This is the ρ parameter for bivariate normal correlation.
                The realized discrete gamma correlation ρ_dG will be somewhat different.
        
        Returns:
            Tuple of (rates, probabilities, transition_matrix)
            - rates: List of rate values for each category
            - probabilities: Stationary probabilities for each category
            - transition_matrix: K×K matrix for correlated rates (empty list if no correlation)
        """
        if invariant_proportion < 0.0 or invariant_proportion >= 1.0:
            raise ValueError(f"invariant_proportion must be in [0, 1), received: {invariant_proportion}")
        
        if site_rate_correlation < 0.0 or site_rate_correlation >= 1.0:
            raise ValueError(f"site_rate_correlation must be in [0, 1), received: {site_rate_correlation}")
        
        # Validate correlation requires multiple categories
        if site_rate_correlation > 0.0 and gamma_categories == 1:
            warnings.warn(
                "site_rate_correlation > 0 requires gamma_categories > 1. "
                "Setting site_rate_correlation to 0.0"
            )
            site_rate_correlation = 0.0
        
        # Create gamma distribution
        gamma_dist = _Sailfish.GammaDistribution(gamma_alpha, gamma_categories)
        rates = list(gamma_dist.getAllRates())
        probs = list(gamma_dist.getAllRatesProb())
        
        # Add invariant sites category if requested
        if invariant_proportion > 0.0:
            # Scale existing probabilities
            scale_factor = 1.0 - invariant_proportion
            probs = [p * scale_factor for p in probs]
            
            # Add invariant category at the beginning
            rates.insert(0, 0.0)
            probs.insert(0, invariant_proportion)
        
        # Build transition matrix for correlated rates
        transition_matrix = []
        if site_rate_correlation > 0.0:
            if invariant_proportion > 0.0:
                warnings.warn(
                    "site_rate_correlation and invariant_sites_proportion cannot be used together. "
                    "Using invariant sites only, ignoring correlation."
                )
            else:
                try:
                    from msasim.correlation import build_auto_gamma_transition_matrix
                    
                    transition_matrix = build_auto_gamma_transition_matrix(
                        alpha=gamma_alpha,
                        categories=gamma_categories,
                        rho=site_rate_correlation
                    )
                    
                except ImportError:
                    warnings.warn(
                        "site_rate_correlation > 0 requires scipy. "
                        "Install with: pip install scipy or pip install 'msasim[correlation]'. "
                        "Ignoring correlation parameter."
                    )
        return rates, probs, transition_matrix
    
    def _init_sub_model(self) -> None:
        self._model_factory = _Sailfish.modelFactory(self._simProtocol._get_Sailfish_tree())
        self._model_factory.set_alphabet(self._alphabet)
        if self._simulation_type == SIMULATION_TYPE.PROTEIN:
            warnings.warn("replacement matrix not provided -> running with default parameters: WAG model")
            self._model_factory.set_replacement_model(_Sailfish.modelCode.WAG)
        else:
            warnings.warn("replacement matrix not provided -> running with default parameters: JC model")
            self._model_factory.set_replacement_model(_Sailfish.modelCode.NUCJC)

    
        rates, probs = self._create_site_rate_model()
        self._model_factory.setSiteRateModel(rates, probs)    



        self._simulator.init_substitution_sim(self._model_factory)
        self._is_sub_model_init = True
    
    def set_replacement_model(
            self,
            model: _Sailfish.modelCode,
            amino_model_file: pathlib.Path = None,
            model_parameters: List = None,
            gamma_parameters_alpha : float = 1.0,
            gamma_parameters_categories: int = 1,
            invariant_sites_proportion: float = 0.0,
            site_rate_correlation: float = 0.0,
        ) -> None:
        if not model:
            raise ValueError(f"please provide a substitution model from the the following list: {_Sailfish.modelCode}")
        if int(gamma_parameters_categories) != gamma_parameters_categories:
            raise ValueError(f"gamma_parameters_catergories has to be a positive int value: received value of {gamma_parameters_categories}")
        self._model_factory = _Sailfish.modelFactory(self._simProtocol._get_Sailfish_tree())

        self._model_factory.set_alphabet(self._alphabet)
        if self._simulation_type == SIMULATION_TYPE.PROTEIN:
            if model_parameters:
                raise ValueError(f"no model parameters are used in protein, recevied value of: {model_parameters}")
            self._model_factory.set_replacement_model(model)
            if model == MODEL_CODES.CUSTOM and amino_model_file:
                self._model_factory.set_amino_replacement_model_file(str(amino_model_file))
        else:
            if model == MODEL_CODES.NUCJC and model_parameters:
                raise ValueError(f"no model parameters in JC model, recevied value of: {model_parameters}")
            self._model_factory.set_replacement_model(model)
            if model == MODEL_CODES.NUCJC and not model_parameters:
                pass
            elif not model_parameters:
                raise ValueError("please provide a model parameters")
            else:
                self._model_factory.set_model_parameters(model_parameters)


        rates, probs, transition_matrix = self._create_site_rate_model(
            gamma_alpha=gamma_parameters_alpha,
            gamma_categories=gamma_parameters_categories,
            invariant_proportion=invariant_sites_proportion,
            site_rate_correlation=site_rate_correlation
        )
        self._model_factory.setSiteRateModel(rates, probs, transition_matrix)

        self._simulator.init_substitution_sim(self._model_factory)

        self._is_sub_model_init = True
    
    def gen_indels(self) -> BlockTreePython:
        return BlockTreePython(self._simulator.gen_indels())
    
    def get_sequences_to_save(self) -> List[bool]:
        return self._simulator.get_saved_nodes_mask()
    
    def save_root_sequence(self):
        self._simulator.save_root_sequence()
    
    def save_all_nodes_sequences(self):
        self._simulator.save_all_nodes_sequences()

    def gen_substitutions(self, length: int):
        if not self._is_sub_model_init:
            self._init_sub_model()
        return self._simulator.gen_substitutions(length)
    
    # @profile
    def simulate(self, times: int = 1) -> List[Msa]:
        Msas = []
        for _ in range(times):
            if self._simProtocol._is_insertion_rate_zero and self._simProtocol._is_deletion_rate_zero:
                msa = Msa(sum(self.get_sequences_to_save()),
                          self._simProtocol.get_sequence_size(),
                          self.get_sequences_to_save())
            else:
                blocktree = self.gen_indels()
                msa = Msa(blocktree._get_Sailfish_blocks(),
                          self._simProtocol._get_root(),
                          self.get_sequences_to_save())

            # sim.init_substitution_sim(mFac)
            if self._simulation_type != SIMULATION_TYPE.NOSUBS:
                substitutions = self.gen_substitutions(msa.get_length())
                msa.fill_substitutions(substitutions)

            Msas.append(msa)
        return Msas
    
    def simulate_low_memory(self, output_file_path: pathlib.Path) -> Msa:
        if self._simProtocol._is_insertion_rate_zero and self._simProtocol._is_deletion_rate_zero:
            msa_length = self._simProtocol.get_sequence_size()
        else:
            blocktree = self.gen_indels()
            msa = Msa(blocktree._get_Sailfish_blocks(),
                        self._simProtocol._get_root(),
                        self.get_sequences_to_save())
            msa_length = msa.get_length()
            self._simulator.set_aligned_sequence_map(msa._msa)

        # sim.init_substitution_sim(mFac)
        if self._simulation_type != SIMULATION_TYPE.NOSUBS:
            self._simulator.gen_substitutions_to_file(msa_length, 
                                                      str(output_file_path))
        else:
            msa.write_msa(str(output_file_path))
    
    def __call__(self) -> Msa:
        return self.simulate(1)[0]
    
    def save_rates(self, is_save: bool) -> None:
        self._simulator.save_site_rates(is_save)
    
    def get_rates(self) -> List[float]:
        return self._simulator.get_site_rates()


"""Multiple sequence alignment output"""

import _Sailfish
from typing import Dict, List

class Msa:
    """MSA result from simulation"""
    
    def __init__(self, species_dict: Dict, root_node, save_list: List[bool]):
        self._msa = _Sailfish.Msa(species_dict, root_node, save_list)

    def generate_msas(self, node):
        self._msa.generate_msas(node)
    
    def get_length(self) -> int:
        return self._msa.length()
    
    def get_num_sequences(self) -> int:
        return self._msa.num_sequences()
    
    def fill_substitutions(self, sequenceContainer) -> None:
        self._msa.fill_substitutions(sequenceContainer)
    
    def print_msa(self) -> str:
        return self._msa.print_msa()
    
    def print_indels(self) -> str:
        return self._msa.print_indels()
    
    def get_msa(self) -> str:
        return self._msa.get_msa_string()
    
    def write_msa(self, file_path) -> None:
        self._msa.write_msa(file_path)



"""
msasim/correlation.py

Auto-discrete-gamma correlation model (Yang 1995) for correlated substitution rates.
"""

import numpy as np
import warnings

try:
    from scipy import stats
    from scipy.integrate import quad
    SCIPY_AVAILABLE = True
except ImportError:
    SCIPY_AVAILABLE = False


def build_auto_gamma_transition_matrix(
    alpha: float,
    categories: int,
    rho: float
) -> np.ndarray:
    """
    Build transition matrix for auto-discrete-gamma model using copula method.
    
    Implements Yang (1995) equations for correlated rate categories.
    Uses bivariate normal copula to generate correlated gamma variates.
    
    Args:
        alpha: Shape parameter of gamma distribution
        categories: Number of discrete rate categories (K)
        rho: Correlation parameter for bivariate normal (-1 to 1)
            This is the input parameter ρ.
            The realized discrete correlation ρ_dG will differ somewhat.
        
    Returns:
        K×K transition matrix M where M[i,j] = P(category j at next site | category i at current site)
        
    Raises:
        ImportError: If scipy is not installed
        ValueError: If parameters are invalid
        
    Reference:
        Yang, Z. (1995). A space-time process model for the evolution of DNA sequences.
        Genetics, 139(2), 993-1005.
    """
    if not SCIPY_AVAILABLE:
        raise ImportError(
            "scipy is required for rate correlation. "
            "Install with: pip install scipy"
        )
    
    if categories < 2:
        raise ValueError(f"Need at least 2 categories, got {categories}")
    if not -1 <= rho <= 1:
        raise ValueError(f"Correlation rho must be in [-1, 1], got {rho}")
    if alpha <= 0:
        raise ValueError(f"Alpha must be positive, got {alpha}")
    
    K = categories
    
    # Handle perfect correlation as special cases (avoid numerical issues)
    if abs(abs(rho) - 1.0) < 1e-6:
        if rho > 0:
            # Perfect positive correlation: X2 = X1 exactly
            # Sites stay in same category: diagonal matrix
            M = np.eye(K)
        else:
            # Perfect negative correlation: X2 = -X1 exactly  
            # Sites flip to opposite category: anti-diagonal matrix
            M = np.eye(K)[:, ::-1]
        return M
    
    # Step 1: Define category boundaries in probability space [0, 1]
    # Each category has probability 1/K
    probs = np.linspace(0, 1, K + 1)
    
    # Step 2: Map to standard normal quantiles
    # For prob p, find z such that Φ(z) = p, where Φ is standard normal CDF
    normal_thresholds = []
    for p in probs:
        if p == 0:
            normal_thresholds.append(-np.inf)
        elif p == 1:
            normal_thresholds.append(np.inf)
        else:
            normal_thresholds.append(stats.norm.ppf(p))
    
    # Step 3: Create bivariate normal with correlation rho
    # scipy handles near-singular cases automatically with allow_singular=True
    mvn = stats.multivariate_normal(
        mean=[0, 0],
        cov=[[1, rho], [rho, 1]],
        allow_singular=True
    )
    
    # Step 4: Compute joint probabilities for each (i,j) rectangle
    joint_probs = np.zeros((K, K))
    
    for i in range(K):
        for j in range(K):
            # Rectangle bounds in standard normal space
            lower_i = normal_thresholds[i]
            upper_i = normal_thresholds[i + 1]
            lower_j = normal_thresholds[j]
            upper_j = normal_thresholds[j + 1]
            
            # Compute P(lower_i < X1 < upper_i, lower_j < X2 < upper_j)
            joint_probs[i, j] = _bivariate_normal_rectangle(
                mvn, lower_i, upper_i, lower_j, upper_j
            )
    
    # Step 5: Normalize to get conditional probabilities M[i,j] = P(j|i)
    # M[i,j] = P(i,j) / P(i) = P(i,j) / (1/K)
    row_sums = joint_probs.sum(axis=1, keepdims=True)
    
    # Handle potential numerical issues
    if np.any(row_sums < 1e-10):
        warnings.warn("Some row sums very small in transition matrix construction")
        row_sums = np.maximum(row_sums, 1e-10)
    
    M = joint_probs / row_sums
    
    return M


def _bivariate_normal_rectangle(
    mvn: stats.multivariate_normal,
    x1_low: float,
    x1_high: float,
    x2_low: float,
    x2_high: float
) -> float:
    """
    Compute P(x1_low < X1 < x1_high, x2_low < X2 < x2_high) for bivariate normal.
    
    Uses CDF differences: P(a < X < b, c < Y < d) = 
        CDF(b,d) - CDF(a,d) - CDF(b,c) + CDF(a,c)
    """
    def safe_cdf(x1, x2):
        """Handle infinite bounds using limits"""
        # Replace infinities with large finite values
        # ±10 gives CDF very close to 0 or 1
        if np.isinf(x1):
            x1 = -10.0 if x1 < 0 else 10.0
        if np.isinf(x2):
            x2 = -10.0 if x2 < 0 else 10.0
        
        try:
            result = mvn.cdf([x1, x2])
            # Handle potential NaN or invalid results
            if not np.isfinite(result):
                # If CDF fails, estimate based on position
                if x1 < -5 or x2 < -5:
                    return 0.0
                elif x1 > 5 or x2 > 5:
                    return 1.0
                else:
                    return 0.5
            return result
        except:
            # Fallback for numerical issues
            if x1 < -5 or x2 < -5:
                return 0.0
            elif x1 > 5 and x2 > 5:
                return 1.0
            else:
                return 0.5
    
    prob = (safe_cdf(x1_high, x2_high) - 
            safe_cdf(x1_low, x2_high) -
            safe_cdf(x1_high, x2_low) + 
            safe_cdf(x1_low, x2_low))
    
    return max(0.0, min(1.0, prob))  # Clamp to [0,1] for numerical safety


def calculate_discrete_gamma_correlation(
    M: np.ndarray,
    alpha: float,
    K: int
) -> float:
    """
    Calculate ρ_dG (discrete gamma correlation) from transition matrix.
    
    Uses Yang (1995) equation (8):
    
    ρ_dG = [Σᵢⱼ (1/K) * Mᵢⱼ * r̄ᵢ * r̄ⱼ - 1] / [Σᵢ (1/K) * r̄ᵢ² - 1]
    
    where r̄ᵢ is the conditional mean rate for category i.
    
    Args:
        M: K×K transition matrix
        alpha: Shape parameter of gamma distribution
        K: Number of categories
        
    Returns:
        ρ_dG: Correlation between rates at adjacent sites
        
    Raises:
        ImportError: If scipy is not installed
    """
    if not SCIPY_AVAILABLE:
        raise ImportError("scipy is required for correlation calculation")
    
    # Compute conditional mean for each category
    r = np.zeros(K)
    
    for i in range(K):
        lower_p = i / K
        upper_p = (i + 1) / K
        
        if lower_p == 0:
            lower_q = 0
        else:
            lower_q = stats.gamma.ppf(lower_p, alpha, scale=1/alpha)
        
        upper_q = stats.gamma.ppf(upper_p, alpha, scale=1/alpha)
        
        def integrand(x):
            return x * stats.gamma.pdf(x, alpha, scale=1/alpha)
        
        numerator, _ = quad(integrand, lower_q, upper_q)
        r[i] = numerator / (1/K)
    
    # Yang's equation (8)
    numerator = 0.0
    for i in range(K):
        for j in range(K):
            numerator += (1/K) * M[i, j] * r[i] * r[j]
    numerator -= 1.0
    
    denominator = np.sum((1/K) * r**2) - 1.0
    
    if abs(denominator) < 1e-10:
        return 0.0  # No variance, no correlation
    
    return numerator / denominator


"""
DEPRECATED: Import from msasim directly instead.

This module exists for backwards compatibility only.
Instead of:
    from msasim import sailfish as sim
    sim.Simulator(...)

Use:
    from msasim import Simulator
    Simulator(...)
"""

import warnings
warnings.warn(
    "Importing from msasim.sailfish is deprecated. "
    "Import directly from msasim instead: 'from msasim import Simulator'",
    DeprecationWarning,
    stacklevel=2
)

# Re-export everything for backwards compatibility
from .distributions import *
from .tree import *
from .protocol import *
from .simulator import *
from .msa import *
from .constants import *



"""Phylogenetic tree handling"""

import _Sailfish
import os
from re import split

def is_newick(tree: str) -> bool:
    """Validate newick format"""
    # from: https://github.com/ila/Newick-validator/blob/master/Newick_Validator.py
    # dividing the string into tokens, to check them singularly
    tokens = split(r'([A-Za-z]+[^A-Za-z,)]+[A-Za-z]+|[0-9.]*[A-Za-z]+[0-9.]+|[0-9.]+\s+[0-9.]+|[0-9.]+|[A-za-z]+|\(|\)|;|:|,)', tree)
    # removing spaces and empty strings (spaces within labels are still present)
    parsed_tokens = list(filter(lambda x: not (x.isspace() or not x), tokens))

    # checking whether the tree ends with ;
    if parsed_tokens[-1] != ';':
        raise ValueError(f"Tree must end with ';'. Got: {tree}")
    return True

class Tree:
    """Phylogenetic tree for simulation"""
    
    def __init__(self, input_str: str):
        is_from_file = os.path.isfile(input_str)
        
        if is_from_file:
            tree_str = open(input_str, 'r').read()
        else:
            tree_str = input_str
            
        if not is_newick(tree_str):
            source = "file" if is_from_file else "string"
            raise ValueError(f"Invalid newick from {source}: {tree_str}")
            
        self._tree = _Sailfish.Tree(input_str, is_from_file)
        self._tree_str = tree_str
    
    def get_num_nodes(self) -> int:
        return self._tree.num_nodes
    
    def get_num_leaves(self) -> int:
        return self._tree.root.num_leaves
    
    def _get_Sailfish_tree(self) -> _Sailfish.Tree:
        """Internal: Get C++ tree object"""
        return self._tree
    
    def __repr__(self) -> str:
        return self._tree_str


"""Distribution classes for indel length modeling"""

import _Sailfish
import math
from typing import List

class Distribution:
    """Base class for discrete distributions"""
    
    def set_dist(self, dist):
        epsilon = 10e-6
        if abs(sum(dist)-1) > epsilon:
            raise ValueError(f"Sum must be 1, got {sum(dist)}")
        for x in dist:
            if x < 0 or x > 1:
                raise ValueError(f"Values must be in [0,1], got {x}")
        self._dist = _Sailfish.DiscreteDistribution(dist)
    
    def _get_Sailfish_dist(self) -> _Sailfish.DiscreteDistribution:
        return self._dist
    

class CustomDistribution(Distribution):
    '''
    Provide a custom discrete distribution to the model.
    '''
    def __init__(self, dist: List[float]):
        self.set_dist(dist)

class GeometricDistribution(Distribution):
    def __init__(self, p: float, truncation: int = 150):
        """
        Calculation of geoemtric moment
        inputs:
        p - p parameter of the geoemtric distribution
        truncation - (optional, by default 150) maximal value of the distribution
        """
        self.p = p
        self.truncation = truncation
        def PMF(x):
            return p*(1-p)**(x-1)
        
        def CDF(x):
            return 1-(1-p)**x
        
        norm_factor = CDF(truncation) - CDF(0)

        probabilities = [PMF(i)/norm_factor for i in range(1, truncation+1)]
        # probabilities = probabilities / norm_factor

        self.set_dist(probabilities)

    def __repr__(self) -> str:
        return f"Geometric distribution: (p={self.p}, truncation{self.truncation})"

class PoissonDistribution(Distribution):
    def __init__(self, p: float, truncation: int = 150):
        """
        Calculation of geoemtric moment
        inputs:
        p - p parameter of the geoemtric distribution
        truncation - (optional, by default 150) maximal value of the distribution
        """
        self.p = p
        self.truncation = truncation
        factorial = math.factorial
        # factorial = lambda z: reduce(operator.mul, [1, 1] if z == 0 else range(1,z+1))

        def PMF(x):
            return ((p**x)*(math.e**-p))*(1.0/factorial(x))
        def CDF(x):
            return (math.e**-p)*sum([(p**i)*(1.0/factorial(i)) for i in range(0,x+1)])

        norm_factor = CDF(truncation) - CDF(0)

        probabilities = [PMF(i)/norm_factor for i in range(1, truncation+1)]

        self.set_dist(probabilities)

    def __repr__(self) -> str:
        return f"Poisson distribution: (p={self.p}, truncation{self.truncation})"

class ZipfDistribution(Distribution):
    def __init__(self, p: float, truncation: int = 150):
        """
        Calculation of geoemtric moment
        inputs:
        p - p parameter of the geoemtric distribution
        truncation - (optional, by default 150) maximal value of the distribution
        """
        self.p = p
        self.truncation = truncation

        norm_factor = sum([(i**-p) for i in range(1,truncation+1)])
        probabilities = [(i**-p)/norm_factor for i in range(1, truncation+1)]

        self.set_dist(probabilities)
    
    def __repr__(self) -> str:
        return f"Zipf distribution: (p={self.p}, truncation{self.truncation})" 




"""msasim - High-performance MSA simulator"""

from .distributions import (
    Distribution,
    CustomDistribution, 
    ZipfDistribution,
    GeometricDistribution,
    PoissonDistribution
)
from .tree import Tree
from .protocol import SimProtocol
from .simulator import Simulator
from .msa import Msa
from .constants import SIMULATION_TYPE, MODEL_CODES

__all__ = [
    'Distribution',
    'CustomDistribution',
    'ZipfDistribution', 
    'GeometricDistribution',
    'PoissonDistribution',
    'Tree',
    'SimProtocol',
    'Simulator',
    'Msa',
    'SIMULATION_TYPE',
    'MODEL_CODES',
]


"""Simulation protocol configuration"""

import _Sailfish
from typing import List, Optional, Dict
from .tree import Tree
from .distributions import Distribution, ZipfDistribution

class SimProtocol:
    """Configuration for MSA simulation"""
    
    def __init__(
        self, 
        tree = None,
        root_seq_size: int = 100,
        deletion_rate: float = 0.0,
        insertion_rate: float = 0.0,
        deletion_dist: Distribution = None,
        insertion_dist: Distribution = None,
        minimum_seq_size: int = 100,
        seed: int = 0,
    ):
        # Handle defaults
        if deletion_dist is None:
            deletion_dist = ZipfDistribution(1.7, 50)
        if insertion_dist is None:
            insertion_dist = ZipfDistribution(1.7, 50)
            
        # Parse tree
        if isinstance(tree, Tree):
            self._tree = tree
        elif isinstance(tree, str):
            self._tree = Tree(tree)
        else:
            raise ValueError("tree must be Tree object or newick string/path")
        
        self._num_branches = self._tree.get_num_nodes() - 1
        self._sim = _Sailfish.SimProtocol(self._tree._get_Sailfish_tree())
        self.set_seed(seed)
        self.set_sequence_size(root_seq_size)
        self._is_deletion_rate_zero = not deletion_rate
        self._is_insertion_rate_zero = not insertion_rate
        self.set_deletion_rates(deletion_rate=deletion_rate)
        self.set_insertion_rates(insertion_rate=insertion_rate)
        self.set_deletion_length_distributions(deletion_dist=deletion_dist)
        self.set_insertion_length_distributions(insertion_dist=insertion_dist)
        self.set_min_sequence_size(min_sequence_size=minimum_seq_size)

    def get_tree(self) -> Tree:
        return self._tree
    
    def _get_Sailfish_tree(self) -> _Sailfish.Tree:
        return self._tree._get_Sailfish_tree()
    
    def _get_root(self):
        return self._tree._get_Sailfish_tree().root
    
    def get_num_branches(self) -> int:
        return self._num_branches
    
    def set_seed(self, seed: int) -> None:
        self._seed = seed
        self._sim.set_seed(seed)
    
    def get_seed(self) -> int:
        return self._seed
    
    def set_sequence_size(self, sequence_size: int) -> None:
        self._sim.set_sequence_size(sequence_size)
        self._root_seq_size = sequence_size
    
    def get_sequence_size(self) -> int:
        return self._root_seq_size
    
    def set_min_sequence_size(self, min_sequence_size: int) -> None:
        self._sim.set_minimum_sequence_size(min_sequence_size)
        self._min_seq_size = min_sequence_size

    
    def set_insertion_rates(self, insertion_rate: Optional[float] = None, insertion_rates: Optional[List[float]] = None) -> None:
        if insertion_rate is not None:
            self.insertion_rates = [insertion_rate] * self._num_branches
            if insertion_rate:
                self._is_insertion_rate_zero = False
        elif insertion_rates:
            if not len(insertion_rates) == self._num_branches:
                raise ValueError(f"The length of the insertaion rates should be equal to the number of branches in the tree. The insertion_rates length is {len(insertion_rates)} and the number of branches is {self._num_branches}. You can pass a single value as insertion_rate which will be used for all branches.")
            self.insertion_rates = insertion_rates
            for insertion_rate in insertion_rates:
                if insertion_rate:
                    self._is_insertion_rate_zero = False
        else:
            raise ValueError("please provide one of the following: insertion_rate (a single value used for all branches), or a insertion_rates (a list of values, each corresponding to a different branch)")
        
        self._sim.set_insertion_rates(self.insertion_rates)
    
    def get_insertion_rate(self, branch_num: int) -> float:
        if branch_num >= self._num_branches:
            raise ValueError(f"The branch number should be between 0 to {self._num_branches} (not included). Received value of {branch_num}")
        return self._sim.get_insertion_rate(branch_num)
    
    def get_all_insertion_rates(self) -> Dict:
        return {i: self.get_insertion_rate(i) for i in range(self._num_branches)}
    
    def set_deletion_rates(self, deletion_rate: Optional[float] = None, deletion_rates: Optional[List[float]] = None) -> None:
        if deletion_rate is not None:
            self.deletion_rates = [deletion_rate] * self._num_branches
            if deletion_rate:
                self._is_deletion_rate_zero = False
        elif deletion_rates:
            if not len(deletion_rates) == self._num_branches:
                raise ValueError(f"The length of the deletion rates should be equal to the number of branches in the tree. The deletion_rates length is {len(deletion_rates)} and the number of branches is {self._num_branches}. You can pass a single value as deletion_rate which will be used for all branches.")
            self.deletion_rates = deletion_rates
            for deletion_rate in deletion_rates:
                if deletion_rate:
                    self._is_deletion_rate_zero = False
        else:
            raise ValueError("please provide one of the following: deletion_rate (a single value used for all branches), or a deletion_rates (a list of values, each corresponding to a different branch)")
        
        self._sim.set_deletion_rates(self.deletion_rates)
    
    def get_deletion_rate(self, branch_num: int) -> float:
        if branch_num >= self._num_branches:
            raise ValueError(f"The branch number should be between 0 to {self._num_branches} (not included). Received value of {branch_num}")
        return self._sim.get_deletion_rate(branch_num)
    
    def get_all_deletion_rates(self) -> Dict:
        return {i: self.get_deletion_rate(i) for i in range(self._num_branches)}
    
    def set_insertion_length_distributions(self, insertion_dist: Optional[Distribution] = None, insertion_dists: Optional[List[Distribution]] = None) -> None:
        if insertion_dist:
            self.insertion_dists = [insertion_dist] * self._num_branches
        elif insertion_dists:
            if not len(insertion_dists) == self._num_branches:
                raise ValueError(f"The length of the insertion dists should be equal to the number of branches in the tree. The insertion_dists length is {len(insertion_dists)} and the number of branches is {self._num_branches}. You can pass a single value as insertion_dist which will be used for all branches.")
            self.insertion_dists = insertion_dists
        else:
            raise ValueError("please provide one of the following: deletion_rate (a single value used for all branches), or a deletion_rates (a list of values, each corresponding to a different branch)")
        
        self._sim.set_insertion_length_distributions([dist._get_Sailfish_dist() for dist in self.insertion_dists])
    
    def get_insertion_length_distribution(self, branch_num: int) -> Distribution:
        if branch_num >= self._num_branches:
            raise ValueError(f"The branch number should be between 0 to {self._num_branches} (not included). Received value of {branch_num}")
        return self.insertion_dists[branch_num]
    
    def get_all_insertion_length_distribution(self) -> Dict:
        return {i: self.get_insertion_length_distribution(i) for i in range(self._num_branches)}
    
    def set_deletion_length_distributions(self, deletion_dist: Optional[Distribution] = None, deletion_dists: Optional[List[Distribution]] = None) -> None:
        if deletion_dist:
            self.deletion_dists = [deletion_dist] * self._num_branches
        elif deletion_dists:
            if not len(deletion_dists) == self._num_branches:
                raise ValueError(f"The length of the deletion dists should be equal to the number of branches in the tree. The deletion_dists length is {len(deletion_dists)} and the number of branches is {self._num_branches}. You can pass a single value as deletion_dist which will be used for all branches.")
            self.deletion_dists = deletion_dists
        else:
            raise ValueError("please provide one of the following: deletion_rate (a single value used for all branches), or a deletion_rates (a list of values, each corresponding to a different branch)")
        
        self._sim.set_deletion_length_distributions([dist._get_Sailfish_dist() for dist in self.deletion_dists])
    
    def get_deletion_length_distribution(self, branch_num: int) -> Distribution:
        if branch_num >= self._num_branches:
            raise ValueError(f"The branch number should be between 0 to {self._num_branches} (not included). Received value of {branch_num}")
        return self.deletion_dists[branch_num]
    
    def get_all_deletion_length_distribution(self) -> Dict:
        return {i: self.get_deletion_length_distribution(i) for i in range(self._num_branches)}
